#!/bin/bash

if [ -z $DODWAN_HOME ]; then
    echo "Error: \$DODWAN_HOME is not defined."
    exit
fi

version=3.1 
classpath="${DODWAN_HOME}/libs/dodwan-${version}.jar"

# ------------------------------------------------------------
init() {

    dodwan_log_dir=/run/shm/${USER}/dodwan
    node_dir=${dodwan_log_dir}/var/node/${node_id}
    pid_file=${node_dir}/pid
    cport_file=${node_dir}/cport
    fifo_file=${node_dir}/in
    
    if [ -e $cport_file ] ;  then
	cport=$(cat ${cport_file})
    fi
}

# ------------------------------------------------------------
send_command() {

#    if [ ! -e ${fifo_file} ] ; then
#	echo "Error: file ${fifo_file} does not exist"
#	exit 1
#    fi
#    echo $* > ${fifo_file}

    if [ -z $cport ] ; then
	echo "Error: no console is available"
	exit 1
    fi
    echo $* | nc -q 1 localhost $cport | sed -e 's/^% //'
}

# ------------------------------------------------------------
console() {

    init

    check_running
    
    if [ -z $cport ] ; then
	echo "Error: no console is available"
	exit 1
    fi
    nc localhost $cport
}

# ------------------------------------------------------------
check_running() {
    
    if [ ! -e ${pid_file} ] ; then
	echo "Error: it seems node ${node_id} is not running"
	exit 1
    fi
}

# ------------------------------------------------------------
start_node() {
    #
    #  Starts a DoDWAN node
    #
    #  Uses the following variables:
    #
    # node_id        : id of the node (required)
    # node_start_time: time when the node should start (EPOCH in ms, optional)
    # node_end_time  : time when the node should stop (EPOCH in ms, optional)
    # node_seed      : seed to be used by that node's random generator (optional)
    # jvm_opts       : options for the JVM (optional)
    # classpath      : class path for Java code
    # dodwan_log_dir : log directory (required)
    # console_port   : TCP port number to be used by the node's console (optional)

    echo Starting node $node_id
    
    init

    if [ -e ${pid_file} ] ; then
	pid=$(cat ${pid_file})
	echo "Error: It seems node ${node_id} is already running (with pid ${pid})"
	echo "       If that is not the case, please remove the pid file with:"
	echo "       % rm ${pid_file}"
	exit 1
    fi
    
    mkdir -p $node_dir

    console_port=0
    props="-Ddodwan.host=${node_id} -Ddodwan.directory=${node_dir} -Ddodwan.base=${dodwan_log_dir}/base"
    if [ ! -z $console_port ] ;  then 
	props="$props -Ddodwan.console_port=${console_port}"
    fi
    
    opts=""
    if [ ! -z $node_start_time ] ;  then
	opts="$opts -begin $node_start_time"
    fi
    if [ ! -z $node_stop_time ] ;  then
	opts="$opts -end $node_stop_time"
    fi
    if [ ! -z $node_seed ] ;  then
	opts="$opts -seed $node_seed"
    fi
    
    # Passing initialization commands to this node
    if [ ! -z "$init_cmd" ] ; then
	echo $init_cmd > ${node_dir}/cmd
	opts="$opts -c ${node_dir}/cmd"
    fi

#    if [ ! -e ${fifo_file} ] ; then
#	mkfifo ${fifo_file}
#    fi
#    opts="$opts -i ${fifo_file}"

    # Starting DoDWAN node
#    java -Xms4m -Xmx4m \
    java ${jvm_opts} ${props} \
	 -cp $classpath \
	 casa.dodwan.run.dodwand ${opts} \
	 >> ${node_dir}/out &
    echo $! > ${pid_file}
    
}

# ------------------------------------------------------------
stop_node() {
    #
    #  Stops a DoDWAN node
    #
    #  Uses the following variables:
    #
    # node_id      : id of the node (required)

    echo Stopping node $node_id
    
    init

    check_running
    
    pid=$(cat $pid_file)
    kill $pid >& /dev/null
#    rm -f $pid_file $cport_file
}

# ------------------------------------------------------------
status() {
    #
    # Shows the status of DoDWAN (i.e. running or not running)
    #
    # node_id: id of the node (required)

    init

    if [ -e ${pid_file} ] ; then
	echo "Node ${node_id} is running"
    else
	echo "Node ${node_id} is not running"
    fi
}

# ------------------------------------------------------------
clear() {
    #
    # Flushes the cache of DoDWAN
    #
    # node_id: id of the node (required)

    init

    if [ -e ${pid_file} ] ; then
	send_command "do ca cl"
    else
	rm -Rf ${node_dir}/cache/*
    fi
}

# ------------------------------------------------------------
publish() {
    #
    # Publish a message, with a file as its payload
    #
    # node_id: id of the node (required)
    # $1     : desc (comma-separated list of name=value pairs)
    # $2     : fname (file to be published)

    desc=$1
    fname=$2
    
    init

    check_running

    if [ ! -e $fname ] ; then
	echo "Error: file $fname does not exist"
	exit 1
    fi

    send_command "do ps p -desc \"src=${node_id},${desc}\" -f $fname"
}

# ------------------------------------------------------------
subscribe() {
    #
    # Sets a subscription
    #
    # node_id: id of the node (required)
    # $1     : key (used to unsubscribe)
    # $2     : pattern (comma-separated list of name=value pairs)
    # $3-*   : options (e.g., -d <dir> | -e <cmd>desc)

    key=$1
    pattern=$2
    shift 2
    options=$*

    init

    check_running

    send_command "do ps add -k ${key} -desc \"${pattern}\" ${options}"
}

# ------------------------------------------------------------
unsubscribe() {
    #
    # Remove subscriptions
    #
    # node_id: id of the node (required)
    # keys   : keys of the subscriptions to be removed

    keys=$*
    
    init

    check_running

    send_command "do ps rem ${keys}"
}
