/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.cache;

import java.io.BufferedReader;
import java.io.File;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Date;
import java.util.BitSet;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.Payload;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;
import casa.dodwan.message.MessageSAR;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.Command;
import casa.dodwan.util.Console;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.Environment;
import casa.dodwan.util.Host;

public class CacheServiceConsole extends Console {
    
    public CacheService cacheService_;
    
    // ------------------------------------------------------------
    public CacheServiceConsole(CacheService cacheService) {
	
	cacheService_ = cacheService;
	prompt_ = "cache% ";
    }
    
    // ============================================================
    // MAIN COMMANDS
    // ============================================================
    
    // ------------------------------------------------------------
    public CacheServiceConsole(String prompt, CacheService cacheService) {
	
	cacheService_ = cacheService;
	prompt_ = prompt;
    }
    
    // ------------------------------------------------------------
    public void cacheDisable(String[] arguments, PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments.length == 1) {
	    cacheService_.disable();
	    return;
	}
	
	for (int i = 1; i < arguments.length; i++) {
	    //  			if (arguments[i].startsWith("p"))
	    //  				cacheService_.environment.disable();
	    
	    // if (arguments[i].startsWith("a"))
	    // 	cacheService_.advertiser.disable();
	    
	    // if (arguments[i].startsWith("f"))
	    // 	cacheService_.setFiltering(false);
	}
    }
    
    // ------------------------------------------------------------
    public void cacheEnable(String[] arguments, PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments.length == 1) {
	    cacheService_.enable();
	    return;
	}
	
	for (int i = 1; i < arguments.length; i++) {
	    //  			if (arguments[i].startsWith("p"))
	    //  				cacheService_.environment.enable();
	    
	    // if (arguments[i].startsWith("a"))
	    // 	cacheService_.advertiser.enable();
	    
	    // if (arguments[i].startsWith("f"))
	    // 	cacheService_.setFiltering(true);
	}
	
    }
    
    // ------------------------------------------------------------
    private String humanReadable(long value) {

	if (value >= 1048576)
	    return String.valueOf(value / 1048576) + "MB";
	if (value >= 1024)
	    return String.valueOf(value / 1024) + "kB";
	return String.valueOf(value) + "B";
    }

    // ------------------------------------------------------------
    public void cacheKeys(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}

	cacheStatus(arguments, out);

	long used    = cacheService_.cache.getWeight();

	long now = System.currentTimeMillis();
	out.println("\nLifetime\tSize\tDocument keys\tFragments\n");
	for (String key : cacheService_.cache.getKeys()) {
	    Descriptor desc = cacheService_.cache.getDescriptor(key);
	    long size = cacheService_.cache.getWeight(key);
	    boolean cancelled = cacheService_.cache.isCancelled(key);
	    BitSet fragments = cacheService_.cache.getFragments(key);
	    used -= size;
	    Date deadline = desc.getDeadline();
	    long lifetime = -1;
	    if (deadline != null)
		lifetime = deadline.getTime() - now;
	    out.println((lifetime == -1?"--:--:--":DataFormatting.to_hh_mm_ss(lifetime)) 
			+ "\t"
			+ (cancelled?"*":size)
			+ "\t"
			+ key
			+ "\t"
			+ fragments.cardinality() + "/" + desc.getNbFragments());
	}

	if (used != 0)
	    out.println("\n\nUh oh! Invalid cache size...");
    }
    
    // ------------------------------------------------------------
    public void cacheLoad(String[] arguments, PrintStream out) {
	
	if (arguments.length <= 1) {
	    help("  ", out);
	    return;
	}
	
	int idx = ArgumentParsing.searchIndex("-flush", arguments);
	
	boolean flush = (idx != -1);
	if (!flush)
	    idx = 0;
	
	for (int i = idx + 1; i < arguments.length; i++) {
	    try {
		String name = arguments[i];
		out.println("Loading the document stored in file " + name);
		cacheService_.load(new File(name), flush);
	    }
	    catch (Exception e) {
		out.println(e);
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void cacheClear(String[] arguments, PrintStream out) {
	
	if (arguments.length > 1) {
	    help("  ", out);
	    return;
	}
	
	try {
	    out.println("Clearing cache");
	    cacheService_.cache.clear();
	}
	catch (Exception e) {
	    out.println("  Failed to clear cache");
	}
    }
    
    // ------------------------------------------------------------
    public void cacheRemove(String[] arguments, PrintStream out) {
	
	if (arguments.length == 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("remove");
	
	for (int i = 1; i < arguments.length; i++) {
	    try {
		cacheService_.cache.remove(arguments[i]);
	    }
	    catch (Exception e) {
		out.println("  Failed to remove document " + arguments[i]
			    + " from cache");
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void cacheCancel(String[] arguments, PrintStream out) {
	
	if (arguments.length == 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("cancel");
	
	for (int i = 1; i < arguments.length; i++) {
	    try {
		cacheService_.cache.cancel(arguments[i]);
	    }
	    catch (Exception e) {
		out.println("  Failed to cancel document " + arguments[i]);
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void cacheResume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	cacheService_.resume();
    }
    
    // ------------------------------------------------------------
    public void cacheSave(String[] arguments, PrintStream out) {
	
	if (arguments.length != 3) {
	    help("  ", out);
	    return;
	}
	
	String key = arguments[1];
	String fname = arguments[2];
	
	try {
	    File file = new File(fname);
	    MessageSAR.getAsFile(key, file, cacheService_.cache);
	}
	catch (Exception e) {
	    out.println("Could not save document in file '" + fname + "'");
	    return;
	}
    }
    
    // ------------------------------------------------------------
    public void cacheSet(String[] arguments, PrintStream out) {
	
	if (arguments.length < 3) {
	    help("  ", out);
	    return;
	}
	
	try{
	    
	    try {
		int capacity = ArgumentParsing.searchInt("-c", arguments);
		DodwanEnvironment.getInstance().cache_capacity = capacity;
//  		((SmartMessageArchiver) cacheService_.archiver)
//  		    .setCapacity(capacity);
	    }
	    catch (ArgumentNotFoundException e) {
	    }
	    
	}
	catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	}
	catch (Exception e) {
	    out.println(e);
	}
    }
    
    // ------------------------------------------------------------
    public void cacheShow(String[] arguments, PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("show");
	
	for (int i = 1; i < arguments.length; i++) {
	    Descriptor desc = 
		cacheService_.cache.getDescriptor(arguments[i]);
	    if (desc != null) {
		out.println("Descriptor: " 
			    + desc);
	    }
	}
    }
    
    // ------------------------------------------------------------
    public void cacheStatus(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	int capacity   = DodwanEnvironment.getInstance().cache_capacity;
	long used      = cacheService_.cache.getWeight();
	int nbElts     = cacheService_.cache.getNbElements();

	if (capacity > 0) {
	    long free = capacity - used;
	    out.println("\nCache: " + nbElts + " msgs, "
			+ humanReadable(capacity) + " total, " 
			+ humanReadable(used) + " used in payloads ("
			+ (100 * used / capacity) + "%), " 
			+ humanReadable(free) + " free ("
			+ (100 * free / capacity) + "%)");
	}
	else
	    out.println("\nCache: " + nbElts + " msgs, "
			+ humanReadable(used) + " used in payloads");

    }
    
    // ------------------------------------------------------------
    public void cacheSuspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("  ", out);
	    return;
	}
	
	cacheService_.suspend();
    }
    
    // ------------------------------------------------------------
    public void cacheVerbosity(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("  ", out);
	    return;
	}
	
	int level = -1;
	try {
	    level = Integer.parseInt(arguments[1]);
	}
	catch (Exception e) {
	    out.println("Could not parse verbosity level");
	}
	
	if (level >= 0)
	    cacheService_.setVerbosity(level);
	;
    }
    
    // ------------------------------------------------------------
    @Override
	public void help(String offset, PrintStream out) {
	out.println(offset
		    + "[ena]ble / [d]isable\n"
		    + offset
		    + "[su]spend / [res]ume\n"
		    + offset
		    + "[st]atus\n"
		    + offset
		    + "[v]erbosity <level>\n"
		    + offset
		    + "[cl]ear\n"
		    + offset
		    + "[k]eys\n"
		    + offset
		    + "[sh]ow <docId>+\n"
		    + offset
		    // + "[c]ancel <-s source> <-d destination>\n"
		    // + offset + "[d]elete <docId>+\n"
		    // + offset
		    + "[ca]ncel <docId>+\n"
		    + offset
		    + "[rem]ove <docId>+\n"
		    + offset
		    + "[se]t <-c capacity>\n"
		    // + offset
		    // + "[fi]lter <-s source> <-d destination>\n"
		    + offset
		    + "[n]eighbourhood [e]nable / [d]isable / [su]spend / [res]ume\n"
		    + offset + "[p]rofile <-d desc>\n"
		    + offset + "[l]oad <-flush> [file|dir]+\n" + offset
		    + "[sa]ve [docId] [file]"); 
		    // + offset + "[a]dvertise");
    }
    
    @Override
	public void processCommand(String[] arguments, BufferedReader reader,
				   PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	// if (arguments[0].startsWith("a")) {
	//     cacheAdvertise(arguments, out);
	//     return;
	// }
	
	if (arguments[0].startsWith("ca")) {
	    cacheCancel(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("cl")) {
	    cacheClear(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("di")) {
	    cacheDisable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("ena")) {
	    cacheEnable(arguments, out);
	    return;
	}
	
	// if (arguments[0].startsWith("f")) {
	//     cacheAddToFilter(arguments, out);
	//     return;
	// }
	
	if (arguments[0].startsWith("k")) {
	    cacheKeys(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("h")) {
	    out.println("Cache Service: Command Set\n");
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("l")) {
	    cacheLoad(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("v")) {
	    cacheVerbosity(arguments, out);
	    return;
	}
	
	// if (arguments[0].startsWith("p")) {
	//     cacheAddToProfile(arguments, out);
	//     return;
	// }
	
	if (arguments[0].startsWith("rem")) {
	    cacheRemove(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("res")) {
	    cacheResume(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("sa")) {
	    cacheSave(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("se")) {
	    cacheSet(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("sh")) {
	    cacheShow(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("st")) {
	    cacheStatus(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("su")) {
	    cacheSuspend(arguments, out);
	    return;
	}
	
	out.println("Unknown command");
	help("  ", out);
    }
    
    // ============================================================
    // COMMAND PARSING
    // ============================================================
    
    // ------------------------------------------------------------
    public void resume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 0) {
	    help("  ", out);
	    return;
	}
	
	cacheService_.resume();
    }
    
    // ------------------------------------------------------------
    public void source(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	parseCommandFile(arguments[0], out);
    }
    
    // ------------------------------------------------------------
    public void suspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 0) {
	    help("  ", out);
	    return;
	}
	
	cacheService_.suspend();
    }
    
}
