/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.docware;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.Keyable;
import casa.dodwan.util.StringGenerator;

/**
 * Descriptor of a DoDWAN message. A descriptor is composed of attributes, each
 * having a name and a value. A descriptor contains internal attributes (with
 * reserved names starting with an underscore character) and user-defined
 * attributes (with freely chosen names, but excluding reserved ones). All
 * attributes can be accessed with a generic getter/setter. For convenience,
 * some of the internal attributes have also specific getters/setters.
 */
public class Descriptor implements Keyable<String> {
    
    /**
     * @undocumented
     */
    private Map<String,String> attributes_ = null;
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Set<Map.Entry<String, String>> getEntrySet() {
	return attributes_.entrySet();
    }

    // ------------------------------------------------------------
    /**
     * Returns the number of attributes in this <code>Descriptor</code>.
     * @return the number of attributes
     */
    public int getNbAttributes() {
	return attributes_.size();
    }

    // ------------------------------------------------------------
    /**
     * Returns the value of the attribute with the specified name in this
     * <code>Descriptor</code>, or <code>null</code> if this attribute does not
     * exist.
     * @param name the name of the attribute whose value should be returned
     * @return the value of this attribute (or <code>null</code> if this
     *         attribute does not exist)
     */
    public String getAttribute(String name) {
	return attributes_.get(name);
    }

    // ------------------------------------------------------------
    /**
     * Returns the names of the attributes defined in this
     * <code>Descriptor</code>.
     * @return the names of the attributes defined in this
     * <code>Descriptor</code>.
     */
    public Set<String> getNames() {
	return attributes_.keySet();
    }

    // ------------------------------------------------------------
    /**
     * Removes the attribute with the specified name from this
     * <code>Descriptor</code>.
     * @param name the name of the attribute that should be removed
     */
    public void removeAttribute(String name) {
	attributes_.remove(name);
    }

    // ------------------------------------------------------------
    /**
     * Sets a new attribute (or changes an existing one) in this
     * <code>Descriptor</code>..
     * @param name the name of the attribute
     * @param value the value assigned to this attribute
     */
    public void setAttribute(String name, String value) {
	attributes_.put(name, value);
    }

   // ------------------------------------------------------------
    /**
     * Creates an empty descriptor.
     */
    public Descriptor() {
	attributes_ = new HashMap<String,String>();
    }
	
    // ------------------------------------------------------------
    /**
     * Creates a descriptor from a string containing a list of attributes.
     * The string should be composed of comma-separated list of attribute assignments 
     * of the form: key=value,key=value,...
     * @param attrs the string containing attributes.
     * @return the descriptor formed from this list of attributes.
     */
    public static Descriptor fromAttributes(String attrs) {
	Descriptor desc = new Descriptor();
	// Extracting descriptor attributes and values
	String[] descAttr = attrs.split(",");
	for (int i = 0; i < descAttr.length; i++) {
	    String[] attr = descAttr[i].split("=");
	    desc.setAttribute(attr[0], attr[1]);
	}
	return desc;
    }

    // ------------------------------------------------------------
    /**
     * Creates a descriptor as a copy of the specified descriptor.
     * @param other the other descriptor that must be copied.
     */
    public Descriptor(Descriptor other) {

	attributes_ = new HashMap<String,String>();
	for (Object obj: other.attributes_.keySet()) {
	    String key = (String)obj;
	    setAttribute(key, other.getAttribute(key));
	}
    }

    // ------------------------------------------------------------
    /**
     * Merges this descriptor with another one. The resulting descriptor
     * contains  all the attributes present in this descriptor, completed by the 
     * attributes of the other descriptor that are not present 
     * in this descriptor.
     * @param other the other descriptor 
     * @return The merged descriptor
     */
    public Descriptor mergeWith(Descriptor other) {
	
	for (Object obj: other.attributes_.keySet()) {
	    String key = (String)obj;
	    if (getAttribute(key) == null)
		setAttribute(key, other.getAttribute(key));
	}
	return this;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setAttribute(String name, List<String> values) {

	setAttribute(name, DataFormatting.toString(values, ","));
    }
	
    // ------------------------------------------------------------
     /**
     * @undocumented
     */
    public void setAttribute(String name, Set<String> values) {

	setAttribute(name, DataFormatting.toString(values, ","));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the value of the ackId field in this <code>Descriptor</code>.
     * @param ackId the value of the "_ackId" field in this
     *        <code>Descriptor</code>
     */
     public void setAckId(String ackId) {
	 setAttribute("_ackid", ackId);
     }

     // ------------------------------------------------------------
     /**
      * @undocumented
      *
      * Returns the value of the ackId field in this <code>Descriptor</code>.
      * @return the value of the "_ackId" field in this <code>Descriptor</code>
      *         (or <code>null</code> if this attribute does not exist)
      */
     public String getAckId() {
	 return getAttribute("_ackid");
     }
     

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the ackRequests field in this
     * <code>Descriptor</code>..
     * @return the value of the "_ackr" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public Set<String> getAckRequests() {

	String str = getAttribute("_ackr");

	return DataFormatting.splitToSet(str, ",");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the value of the "_ackr" field in this <code>Descriptor</code>.
     * @param ackRequests the value of the "_ackr" field in this
     *        <code>Descriptor</code>
     */
    public void setAckRequests(Set<String> ackRequests) {

	String str = DataFormatting.toString(ackRequests, ",");
	setAttribute("_ackr", str);
    }

    // ------------------------------------------------------------
    /**
     * Returns the date of this descriptor. This date is the date of 
     * emission of the message outlined by this descriptor.
     * @return The value of the "_date" attribute, or <code>null</code> if this 
     *         attribute does not exist.
     */
    public Date getDate() {

	String str = getAttribute("_date");
	try {
	    return DataFormatting.from_date(str);
	}
	catch (Exception e) {
	    return null;
	}
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the emission date (or changes the existing one) in this descriptor. 
     * @param date the new value of the "_date" attribute.
     */
    public void setDate(Date date) {

	setAttribute("_date", DataFormatting.to_date(date));
    }

    // ------------------------------------------------------------
    /**
     * Returns the deadline of this descriptor.  This is the date at
     * which the message should be forgotten by any host that holds a
     * copy of it. Hence the deadline marks the termination of the
     * message's dissemination.
     * @return the value of the "_dline" attribute, 
     *         or <code>null</code> if this attribute does not exist.
     */
    public Date getDeadline() {

	String str = getAttribute("_dline");
	if (str == null)
	    return null;

	try {
	    Date result = DataFormatting.from_date(str);
	    return result;
	}
	catch (Exception e) {
	    System.err.println("getDeadline(): invalid value");
	    return null;
	}
    }

    // ------------------------------------------------------------
    /**
     * Sets the deadline (or changes the existing one) in this descriptor.
     * @param deadline the new value of the "_dline" attribute.
     */
    public void setDeadline(Date deadline) {
	if (deadline != null)
	    setAttribute("_dline", DataFormatting.to_date(deadline));
	else
	    removeAttribute("_dline");
    }

    // ------------------------------------------------------------
    /**
     * Returns the destination of this descriptor. 
     * 
     * @return the value of the "_dest" attribute, 
     *         or <code>null</code> if this attribute does not exist.
     */
    public String getDestination() {
	return getAttribute("_dest");
    }

    // ------------------------------------------------------------
    /**
     * Sets the destination (or changes the existing one) in this descriptor.
     * @param destination the new value of the "_dest" attribute
     */
    public void setDestination(String destination) {
	setAttribute("_dest", destination);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Remove the destination attribute (if any) from this descriptor.
     */
    public void removeDestination() {
	removeAttribute("_dest");
    }

    // ------------------------------------------------------------
    /**
     * Returns the document identifier of this descriptor.
     * @return the value of the "_docid" attribute.
     *         or <code>null</code> if this attribute does not exist.
     */
    public String getDocumentId() {
	return getAttribute("_docid");
    }

    // ------------------------------------------------------------
    /**
     * Returns the key that can serve as a unique identifier for this
     * descriptor.  This method is equivalent to the <code>getDocumentId</code> method. 
     * The value returned is the value of the "_docid" attribute.
     * @return the value of the "_docid" attribute.
     *         or <code>null</code> if this attribute does not exist.
     */
    public String getKey() {
	return getDocumentId();
    }

    // ------------------------------------------------------------
    /**
     * Sets the document id (or changes the existing one) in this descriptor.
     * @param documentId the new value of the "_docid" attribute.
     */
    public void setDocumentId(String documentId) {
	setAttribute("_docid", documentId);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public List<String> getForwarders() {
	
	String str = getAttribute("_fwd");
	return DataFormatting.splitToList(str, ",");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setForwarders(List<String> forwarders) {

	setAttribute("_fwd", forwarders);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Set<String> getPeers() {
	
	String str = getAttribute("_ngh");
	return DataFormatting.splitToSet(str, ",");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setPeers(Set<String> peers) {

	setAttribute("_ngh", DataFormatting.toString(peers, ","));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the forwarderType field in this
     * <code>Descriptor</code>..
     * @return the value of the "_ftype" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public String getForwarderType() {
	return getAttribute("_ftype");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setForwarderType(String type) {
	setAttribute("_ftype", type);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the fragmentId field in this
     * <code>Descriptor</code>..
     * @return the value of the "_frid" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public int getFragmentId() {

	String str = getAttribute("_frid");

	int result;
	try {
	    result = Integer.parseInt(str);
	}
	catch (Exception e) {
	    result = -1;
	}
	return result;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the fragment id (or changes the existing one) in this descriptor
     * @param fragmentId the new value of the "_frid" attribute.
     */
    public void setFragmentId(int fragmentId) {
	setAttribute("_frid", String.valueOf(fragmentId));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the hopsToLive field in this
     * <code>Descriptor</code>..
     * @return the value of the "_htl" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public int getHopsToLive() {

	String str = getAttribute("_htl");

	int result;
	try {
	    result = Integer.parseInt(str);
	}
	catch (Exception e) {
	    result = 0;
	}
	return result;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the number of hops to live (or changes the existing one) in this descriptor.
     * @param hopsToLive the new value of the "_htl" attribute.
     */
    public void setHopsToLive(int hopsToLive) {
	setAttribute("_htl", String.valueOf(hopsToLive));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public String getLastRelay() {
	return getAttribute("_lrel");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setLastRelay(String lastRelay) {
	setAttribute("_lrel", lastRelay);
    }

    // ------------------------------------------------------------
    /**
     * Returns the lifetime of the message outlined by this descriptor.
     * The lifetime is the period of time, expressed in milliseconds, between 
     * the emission date and the deadline.
     * @return the value of the "_ltime" attribute,
     *         or <code>null</code> if this attribute does not exist.
     */
    public long getLifetime() {
	
	String str = getAttribute("_ltime");

	try {
	    return DataFormatting.from_hh_mm_ss(str);
	}
	catch (Exception e) {
	    return -1;
	}
    }

    // ------------------------------------------------------------
    /**
     * Sets the lifetime, expressed in milliseconds, in this descriptor.
     * @param lifetime the new value of the "_ltime" attribute
     */
    public void setLifetime(long lifetime) {

	setAttribute("_ltime", DataFormatting.to_hh_mm_ss(lifetime));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Descriptor getMatchingAttributes(Descriptor regex)
	throws Exception {

	Descriptor result = new Descriptor();
	
	for (Object obj: regex.attributes_.keySet()) {
	    String key = (String)obj;
	    String localValue = getAttribute(key);
	    if (localValue != null) {
		String regexValue = regex.getAttribute(key);
		if (Pattern.matches(regexValue, localValue))
		    result.setAttribute(key, localValue);
	    }
	}
	return ((result.getNbAttributes() > 0)?result:null);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Set<Descriptor> getMatchingDescriptors(Collection<Descriptor> descs) {

	Set<Descriptor> result = new HashSet<Descriptor>();

	for (Descriptor elt : descs) {
	    try {
		if (elt.matches(this))
		    result.add(elt);
	    }
	    catch (Exception e) {
	    }
	}
	return result;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
   public Set<String> getMatchingKeys(Collection<Descriptor> descs) {

	Set<String> result = new HashSet<String>();

	//  		System.out.println("Testing pattern " + this + " against:");
	for (Descriptor elt : descs) {
	    try {
		boolean match = elt.matches(this);
		//  			System.out.println("  " + elt + " -> " + match);
		if (match)
		    result.add(elt.getKey());
	    }
	    catch (Exception e) {
	    }
	}
	return result;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the nbFragments field in this
     * <code>Descriptor</code>..
     * @return the value of the "_nbfrags" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public int getNbFragments() {

		String str = getAttribute("_nbfrags");
		
		int result;
		try {
			result = Integer.parseInt(str);
		}
		catch (Exception e) {
			result = -1;
		}
		return result;
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the number of fragments in this descriptor.
     * @param nbFragments the new value of the "_nbfrags" attribute.
     */
    public void setNbFragments(int nbFragments) {

	setAttribute("_nbfrags", String.valueOf(nbFragments));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the period" field in this <code>Descriptor</code>..
     * @return the value of the "_period" field in this <code>Descriptor</code>
     *         (or <code>-1</code> if this attribute does not exist)
     */
    public long getPeriod() {

	String str = getAttribute("_period");

	try {
	    return DataFormatting.from_hh_mm_ss(str);
	}
	catch (Exception e) {
	    return -1;
	}
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the period in this descriptor.
     * @param period the new value of the "_period" attribute.
     */
    public void setPeriod(long period) {
	setAttribute("_period", DataFormatting.to_hh_mm_ss(period));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public List<String> getRelays() {

	String str = getAttribute("_relays");
	return DataFormatting.splitToList(str, ",");
    }

    // ------------------------------------------------------------
     /**
     * @undocumented
     */
    public void setRelays(List<String> relays) {

	String str = DataFormatting.toString(relays, ",");
	setAttribute("_relays", str);
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the requestType field in this
     * <code>Descriptor</code>..
     * @return the value of the "_rtype" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public String getRequestType() {
	return getAttribute("_rtype");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the value of the requestType field in this <code>Descriptor</code>
     * ..
     * @param requestType the value of the "_rtype" field in this
     *        <code>Descriptor</code>
     */
    public void setRequestType(String requestType) {
	setAttribute("_rtype", requestType);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Returns the value of the serviceType field in this
     * <code>Descriptor</code>..
     * @return the value of the "_stype" field in this
     *         <code>Descriptor</code> (or <code>null</code> if this attribute
     *         does not exist)
     */
    public String getServiceType() {
	return getAttribute("_stype");
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the value of the serviceType field in this <code>Descriptor</code>
     * ..
     * @param type the value of the "_stype" field in this
     *        <code>Descriptor</code>
     */
    public void setServiceType(String serviceType) {
	setAttribute("_stype", serviceType);
    }
    
    // ------------------------------------------------------------
    /**
     * Returns the source of the message outlined by this descriptor.
     * @return the value of the "_src" attribute,
     *         or <code>null</code> if this attribute does not exist.
     */
    public String getSource() {
	return getAttribute("_src");
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     *
     * Sets the source in this descriptor.
     * @param source the new value of the "_src" attribute
     */
    public void setSource(String source) {
	setAttribute("_src", source);
    }
    
    // ------------------------------------------------------------
    /**
     * Tests if this descriptor is equal to another descriptor. They are equal 
     * if all the attributes (name and value) they contains are equal.
     * @param other The other descriptor to compare with
     * @return True if this descriptor is equal to the other descriptor, false otherwise.
     */
    public boolean equals(Descriptor other) {

	if (attributes_.size() != other.attributes_.size())
	    return false;

	for (Object obj: attributes_.keySet()) {
	    String key = (String)obj;
	    if (! getAttribute(key).equals(other.getAttribute(key)))
		return false;
	}

	return true;
    }

    // ------------------------------------------------------------
    /**
     * Matches all the attributes in this descriptor against the
     * attributes contained in the specified descriptor.
     * <p>
     * Attribute matching is performed assuming that the attributes defined in
     * <code>regex</code> define Java regular expressions, against which the
     * attributes defined in this descriptor should be matched.
     * This descriptor is said to "match" <code>regex</code> if,
     * for any attribute defined in <code>regex</code>, there exists an
     * attribute with the same name in this descriptor whose value
     * matches that of the <code>regex</code>'s attribute.
     * @param regex the descriptor whose attributes define
     *        regular expressions that must be matched by the attributes of this
     *        descriptor.
     * @return <code>True</code> if all attributes in <code>regex</code> are
     *         matched by attributes in this descriptor, false
     *         otherwise.
     */
    public boolean matches(Descriptor regex) {

	for (Object obj: regex.attributes_.keySet()) {
	    String key = (String)obj;
	    String localValue = getAttribute(key);
	    if (localValue == null)
		return false;
	    String regexValue = regex.getAttribute(key);
	    if (! Pattern.matches(regexValue, localValue))
		return false;
	}
	return true;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setForwarders(List<String> forwarders, String type) {
	setForwarders(forwarders);
	setForwarderType(type);
    }
    
    // ------------------------------------------------------------
     /**
     * Sets a unique identifier (or change the existing one) in this
     * descriptor. The identifier is chosen arbitrarily.
     */
    public void setUniqueDocumentId() {
	StringGenerator strGen = StringGenerator.getGenerator();
	String docId = strGen.newString();
	setDocumentId(docId);
    }

    // ------------------------------------------------------------
    /**
     *
     * Produces a XML-like <code>String</code> representation of this
     * descriptor.
     * @return a <code>String</code> representation of this
     *         descriptor.
     */
    @Override
	public String toString() {

	String result = "<descriptor ";
	boolean first = true;
	for (String name: attributes_.keySet()) {
	    if (first) 
		first = false;
	    else
		result += ",";
	    String value = attributes_.get(name);
	    result += name + "=\"" + value + "\"";
	}
	result += ">";

	return result;
    }


    // ============================================================
    /**
     * @undocumented
     */
    public static void main(String[] args) throws Exception {

	Set<Descriptor> set = new HashSet<Descriptor>();
	
	Descriptor d1 = new Descriptor();
	d1.setAttribute("_docid", "d1");
	d1.setAttribute("att1", "fg");
	d1.setAttribute("att2", "pg mg");
	set.add(d1);
	System.out.println("d1=" + d1);

	Descriptor d2 = new Descriptor();
	d2.setAttribute("_docid", "d2");
	d2.setAttribute("att1", "ym");
	d2.setAttribute("att2", "pm mm");
	d2.setAttribute("att3", "mm sm");
	set.add(d2);
	System.out.println("d2=" + d2);

	Descriptor d3 = new Descriptor();
	d3.setAttribute("_docid", "d3");
	d3.setAttribute("att1", "nls");
	d3.setAttribute("att2", "pls mls");
	d3.setAttribute("att3", "als mls");
	set.add(d3);
	System.out.println("d3=" + d3);

	Descriptor p1 = new Descriptor();
	p1.setAttribute("att1", "(f|y).*");
	p1.setAttribute("att2", ".*(p|m).*");
	System.out.println("p1=" + p1);

	System.out.println("d1.matches(p1)=" + d1.matches(p1));
	System.out.println("d2.matches(p1)=" + d2.matches(p1));
	System.out.println("d3.matches(p1)=" + d3.matches(p1));

	System.out.println("Matching(d1, p1): " 
			   + d1.getMatchingAttributes(p1));
	System.out.println("Matching(d2, p1): " 
			   + d2.getMatchingAttributes(p1));
	System.out.println("Matching(d3, p1): " 
			   + d3.getMatchingAttributes(p1));

	System.out.println("matchingKeys: " + p1.getMatchingKeys(set));
    }


}
