/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.hub;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;

import casa.dodwan.cache.CacheService;
import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.message.Message;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.util.Console;
import casa.dodwan.util.Logger;
import casa.dodwan.util.Service;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Environment;
import casa.dodwan.util.Domain;

public class Bridge extends Service {

    private Map<String,Domain> domains_ =
	new HashMap<String,Domain>();
    
    public CacheService cacheService;

    public Environment environment;

    /**
     * @undocumented
     */
    private Console console_ = null;
    
    private Object lock_ = new Object();

    private boolean isOnline_ = false;

    private Descriptor pattern = new Descriptor();
    
    // --------------------------------------------------------------------
    /**
     *  
     */
    public Bridge()
	throws Exception {

   	DodwanEnvironment env = DodwanEnvironment.getInstance();
	
	// Activity monitoring --------------------------
	long activityPeriod = env.activity_period;
	
	if (activityPeriod != 0) {
	    
	    File dir = new File(env.directory, "log");
	    String fname = SystemEnvironment.host + "_activity.log";
	    File logFile = new File(dir, fname);
	    
	    casa.dodwan.util.ActivityMonitor al = new casa.dodwan.util.ActivityMonitor(activityPeriod, logFile);
	    al.start();
	}
	
	// -----------------------------------------------
	
	resetLogger();
	
	init(SystemEnvironment.host, env.directory);
    }

    // -------------------------------------------------------------------
    /**
     *  @undocumented
     *  Starts logging DoDWAN events in a new file. 
     */
    public static void resetLogger() {
	
	// System.out.println("Dodwan.resetLogger");

	File dir = new File(DodwanEnvironment.getInstance().directory, "log");
	
	SimpleDateFormat dateFormat = 
	    new SimpleDateFormat("yyyy-MM-dd-HH-mm", Locale.ENGLISH);
	long now = System.currentTimeMillis();
	String fname = SystemEnvironment.host 
	    + "_" + dateFormat.format(now) + ".log";
	    
	File logFile = new File(dir, fname);

	Logger.enable(logFile, true, 
		      SystemEnvironment.host);

	System.out.println("Now logging in file "
			   + logFile);
    }

    // ---------------------------------------------------------------------
    /**
     * @undocumented
     */
    public Console console() {
	if (console_ == null)
	    console_ = new BridgeConsole(this);
	return console_;
    }
 
   
    // ----------------------------------------------------------------------
    private void init(String host, 
		      File dir)
	throws Exception {

	// Set default StringGenerator so that all strings generated
	// will use the local host name as a prefix
	casa.dodwan.util.StringGenerator.setSeed(host);

	cacheService = 
	    new CacheService((dir == null ? null : new File(dir, "cache")));

	enable();

	console_ = new BridgeConsole(this);

	resume();
    }

    // ------------------------------------------------------------
    public void addDomain(String id) {

	if (domains_.containsKey(id)) {
	    Logger.log("bridge", "warning: domain " + id
		       + " already exists");
	    return;
	}

	Domain domain = null;
	try {
	    Logger.log("bridge", "add domain " + id);
	    domain = new Domain(cacheService);
	    domain.gossipingService.addToLocalProfile(pattern);
	}
	catch (Exception e) {
	    System.out.println("Failed to create domain");
	    return;
	}
	console_.addSubConsole(id, "[" + id + "]",
			       domain.console());

	domains_.put(id, domain);
	if (! isSuspended())
	    domain.resume();
    }
    
    // -----------------------------------------------------------------------
    public void removeDomain(String id) {

	if (! domains_.containsKey(id)) {
	    Logger.log("bridge", "warning: domain " + id
		       + " does not exist");
	    return;
	}

	Logger.log("bridge", "remove domain " + id);
	Domain domain = domains_.get(id);
	domain.suspend();
	console_.removeSubConsole(id);
	domains_.remove(id);
    }

    // ------------------------------------------------------------
    public Set<String> getDomains() {

	return domains_.keySet();
    }
    
    // ------------------------------------------------------------
    public Domain getDomain(String id) {

	return domains_.get(id);
    }
    
    // ------------------------------------------------------------
    public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("Bridge.resume()");
	
	super.resume();

	if (! Logger.isEnabled())
	    resetLogger();
	
	Logger.log("bridge", "resume");
    }
    
    // ----------------------------------------------------------------------
    public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("Bridge.suspend()");
	
	super.suspend();

	cacheService.suspend();
	for (Domain dom: domains_.values())
	    dom.suspend();

	if (Logger.isEnabled()) {
	    Logger.log("bridge", "suspend");
	    Logger.disable();
	}
    }


    
    // -----------------------------------------------------------------------
    /**
     * Returns the current status (online/offline) of DoDWAN.
     */
	public boolean isOnline() {
	    return isOnline_;
	}
    
    // -----------------------------------------------------------------------
    public void online() {
	
	if (isOnline_)
	    return;
	
	if (verbosity.isEnabled())
	    System.out.println("Bridge.online()");
	
	Logger.log("bridge", "online");

	for (Domain dom: domains_.values())
	    dom.resume();
	cacheService.resume();

	isOnline_ = true;
    }
    
    // ----------------------------------------------------------------------
    public void offline() {
	
	if (! isOnline_)
	    return;

	if (verbosity.isEnabled())
	    System.out.println("Bridge.offline()");
	
	cacheService.suspend();
	for (Domain dom: domains_.values())
	    dom.suspend();

	Logger.log("bridge", "offline");
	isOnline_ = false;
    }



    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void doWait(){
	synchronized(lock_){
	    try {
		lock_.wait();
	    } 
	    catch(InterruptedException e) {
		System.err.println("Bridge.doWait() failed");
	    }
	}
    }



    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void doNotify(){
	synchronized(lock_){
	    lock_.notify();
	}
    }

 
}
