/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.net;

import java.util.Set;
import java.util.HashSet;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;

import casa.dodwan.util.Logger;
import casa.dodwan.util.Suspendable;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Timer;
import casa.dodwan.util.TimerTask;
import casa.dodwan.util.TaskTrigger;

public class BeaconSender extends Suspendable
    implements Runnable {

    private Message localAnnounce_;
    private String localHost_;
    private long lastAnnounce_;
    
    private TimerTask task_ = null;
    private Timer timer_ = Timer.getInstance();
    private long period_ = 0;
    
    private Channel channel_;
    
    // ------------------------------------------------------------
    public BeaconSender(Channel channel,
			long period) {
	
	channel_ = channel;
	localHost_ = SystemEnvironment.host;
	period_ = period;
	
	// Let us create a template message for the periodic announce
	localAnnounce_ = new Message();
	Descriptor aDesc = new Descriptor();
	aDesc.setAttribute("_btype", "hello");
	aDesc.setAttribute("_sdr", localHost_);
	localAnnounce_.setDescriptor(aDesc);
	localAnnounce_.setPayload(new byte[0]);

	Logger.log("beacon", "channel=" + channel_
		   + ",period=" + period);
    }
    
    // ============================================================
    // Management of the local announce
    // ============================================================
    
    // ------------------------------------------------------------
    public long period() {

	return period_;
    }

    // ------------------------------------------------------------
    public void setPeriod(long period) {
	
	period_ = period;
    }
    
    // ------------------------------------------------------------
    public void setAttribute(String key, String value) {

	Descriptor desc = localAnnounce_.getDescriptor();
	desc.setAttribute(key, value);
    }

    // ------------------------------------------------------------
    private void updateAnnounceMessage() {
	
	Descriptor desc = localAnnounce_.getDescriptor();
	desc.setUniqueDocumentId();
	
	lastAnnounce_ = System.currentTimeMillis();
    }
    
    // ------------------------------------------------------------
    public synchronized void sendAnnounce(String rcv) {
	
	if (isSuspended_)
	    return;

	Descriptor desc = localAnnounce_.getDescriptor();
	
	Logger.log("beacon", "> (" + desc.getAttribute("_btype")
		   + ")"
		   + (rcv!=null?" rcv=" + rcv:""));
	
	try {
	    if (rcv != null) 
		// Sending an announce to one host specifically
		desc.setAttribute("_rcv", rcv);
	    else
		// Sending an announce to any host (possibly all)
		desc.removeAttribute("_rcv");
	    channel_.write(localAnnounce_);
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }

    // ============================================================
    // Implementation of the Runnable interface
    // ============================================================
    
    public void run() {
	
	// updateChannels();
	updateAnnounceMessage();
	
	sendAnnounce(null);
	
    }
    
    // ------------------------------------------------------------
    private void schedule(long delay, long period) {
	
	if (task_ != null)
	    task_.cancel();
	
	task_ = new TaskTrigger(this);
	timer_.schedule(task_, delay, period);
    }
    
    // ------------------------------------------------------------
    public void proceed() {
	
	run();

	if (period_ != 0)
	    schedule(period_, period_);
    }

    // ============================================================
    // Implementation of the Suspendable interface
    // ============================================================

    public void suspend() {
	
	if (isSuspended_) {
	    return;
	}

	localAnnounce_.getDescriptor().setAttribute("_btype", "bye");
	sendAnnounce(null);
	localAnnounce_.getDescriptor().setAttribute("_btype", "hello");

	super.suspend();
	if (task_ != null) {
	    task_.cancel();
	    task_ = null;
	}

	Logger.log("beacon", "suspend");
    }
    
    // ------------------------------------------------------------
    public void resume() {
	
	Logger.log("beacon", "resume");

	super.resume();

	if (period_ != 0) 
	    proceed();
    }
}
