/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.pubsub;

import java.io.File;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import casa.dodwan.gossip.GossipingService;
import casa.dodwan.cache.CacheService;
import casa.dodwan.cache.CacheListener;
import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.Payload;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageReceiver;
import casa.dodwan.message.MessageSAR;
import casa.dodwan.util.Console;
import casa.dodwan.util.Dispatcher;
import casa.dodwan.util.Processor;
import casa.dodwan.util.Service;
import casa.dodwan.util.Sink;

/**
 * DoDWAN publication/subscription of messages. DoDWAN automatically creates a
 * unique instance of this class that should be accessed via a
 * {@link casa.dodwan.run.Dodwan} object. Publication consists in depositing a
 * message in the DoDWAN cache so that this message disseminates in the network.
 * Subscriptions are registered with DoDWAN specifying a pattern of subscription
 * (descriptor that may contain regular expressions) and a handler (a
 * {@link casa.dodwan.run.Processor} object) that will process received messages
 * that match the given pattern. A subscription is identified by a key that can
 * be used for unregistering this subscription. Subscriptions are not
 * persistent: it is the responsibility of the application to save subscriptions
 * on a permanent storage support before stopping DoDWAN and to re-add them at
 * startup if necessary. When a message is received, the processor is called to
 * which is passed the descriptor of the message. The processor can use the key
 * attribute of the descriptor to retrieve the payload of the message (as a byte
 * array or a file).
 */
public class PubSubService 
    extends Service
    implements CacheListener {
    
    private Console console_ = null;
    
    private CacheService cacheService_;

    private GossipingService gossipingService_;

    /**
     * regroupe tous les recepteurs destines a traiter selectivement les
     * documents recus.
     */
    private Dispatcher<Descriptor> receivers_ = new Dispatcher<Descriptor>();
    
    private Map<String, MessageReceiver> map_ = 
	new HashMap<String, MessageReceiver>();

//      private Processor<AssemblyEvent> listener_ = null;
//      private PubSubListener listener_ = null;

    // FG: la gestion d'evenements destinee aux listeners du service
    // PubSub est sans doute superflue. Elle ne sert apparemment qu'au
    // ReceptionHistory, qui pourrait etre invoque directement.
    private PubSubEventDispatcher eventDispatcher_ = 
	new PubSubEventDispatcher();

    private ReceptionHistory history_ = null;

    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public PubSubService(CacheService cacheServ, 
			 GossipingService gossipingServ,
			 File path)
	throws Exception {
	
	super.suspend();

	gossipingService_ = gossipingServ;
	cacheService_ = cacheServ;
	cacheService_.cache.addListener(this);

	if (path != null) {
	    if (!path.exists())
		path.mkdirs();

	    history_ = new ReceptionHistory(new File(path, "history"));
	}
    }
    
    // ------------------------------------------------------------
    /**
     * Adds an event listener. Events pertain to publication, subscription and
     * reception of messages.
     * @param listener the listener to be added.
     */
    public void addListener(PubSubListener listener) {

    	eventDispatcher_.addListener(listener);
    }

    // ------------------------------------------------------------
    /**
     * Removes an event listener. 
     * @param listener the listener to be removed.
     */
     public void removeListener(PubSubListener listener) {

	eventDispatcher_.removeListener(listener);
    }

    // ------------------------------------------------------------
    /**
     * Adds a subscription. 
     * @param key an identifier for the subscription. 
     *        This identifier can be used for subscription removing.
     * @param pattern a descriptor serving as a pattern for content-based reception.
     * @param processor processor for incoming messages (those whose descriptor 
     *        matches the specified pattern). Method <code>process</code> will 
     *        be invoked on this processor when a message is received.
     */
    public void addSubscription(String key, Descriptor pattern, 
				Processor<Descriptor> processor) {

//  	System.out.println("PubSubService.addSubscription(): POINT 0");
	
	if (map_.containsKey(key)){
	    System.out.println("The key " + key + ", is already registered.");
	    return;
	}

	if (verbosity.isEnabled())
	    System.out.println("PubSubService.addSubscription(" 
			       + key + ", "
			       + pattern + ", "
			       + processor + ")");
	
//  	System.out.println("PubSubService.addSubscription(): POINT 1");

	try {
	    // Adjust the local profile based on the specified pattern
	    gossipingService_.addToLocalProfile(pattern);
	    
//  	    System.out.println("PubSubService.addSubscription(): POINT 2");
	    
	    
	    // If a processor has been specified for this pattern, then create
	    // a MessageReceiver accordingly, and check if messages matching
	    // the pattern are already available in the local cache
//             if (processor != null) {
		MessageReceiver receiver = 
		    new MessageReceiver(pattern, processor);
					// cacheService_.cache);
		map_.put(key, receiver);
		
		receivers_.addSink(receiver);
		
//  		System.out.println("PubSubService.addSubscription(): POINT 3");
		
		// Look for documents matching the new pattern in the
		// local cache, and if such documents exist and are
		// complete (i.e., all fragments available), then
		// process them immediately
		Collection<String> keys =
		    cacheService_.cache.getKeys(pattern);
		for (String k : keys) {
		    Descriptor desc = cacheService_.cache.getDescriptor(k);
		    if (desc != null) {
			if (! history_.contains(k)) {
			    if (cacheService_.cache.isComplete(k))
				receiver.write(desc);
			}
			//  			 else
			//  			     System.out.println("PubSubService.addSubscription(): ignoring document " + docId + " that has already been received in the past");
		    }
		}
//	    }
	}
	catch (Exception e) {
	    System.out.println(e);
	}

//  	System.out.println("PubSubService.addSubscription(): POINT 4");
    }


    // ------------------------------------------------------------
    /**
     * Clears all subscriptions.
     */
    public void clear() {

	Set<String> keys = new HashSet<String>();
	keys.addAll(map_.keySet());
	for (String key: keys)
	    removeSubscription(key);
    }
    
    // ------------------------------------------------------------
    /**
     * Removes a subscription. Incoming messages covered by the specified 
     * subscription will no longer be delivered to the subscription's processor. 
     * Does nothing if the identifier <code>key</code> does not correspond to a 
     * registered subscription.
     * @param key the identifier of the subscription to be removed
     */
    public void removeSubscription(String key) {
	
	if (verbosity.isEnabled())
	     System.out.println("PubSubService.removeSubscription("
				+ key + ")");

	 MessageReceiver recv = map_.get(key);
	 if (recv == null)
	     return;

	 Descriptor pattern = recv.getPattern();
	 gossipingService_.removeFromLocalProfile(pattern);

	 receivers_.removeSink(recv);

 //  	recv.removeProcessor(sub.processor());

	 map_.remove(key);
     }

     // ------------------------------------------------------------
     /**
     * Publishes a message. Does nothing if the size of the payload of
     * the specified message is greater than the maximum payload size
     * (see {@link DodwanEnvironment}).
     * @param desc the descriptor of the message to be published.
     * @param buffer the payload of the message to be published.
     * @throws Exception if the publication fails.
     */
    public void publish(Descriptor desc, byte[] buffer) throws Exception {

	if (verbosity.isEnabled())
	    System.out.println("PubSubService.publish()");
	
	if (isSuspended()) {
	    if (verbosity.isEnabled())
		System.out.println("  This service is suspended. Aborting...");
	    return;
	}

	// If no buffer was provided, then let us send an empty buffer
	// instead
	if (buffer == null)
	    buffer = new byte[0];

	int max_payload = DodwanEnvironment.getInstance().max_payload_size;
	if ((max_payload > 0) 
	    && (buffer.length > max_payload)) {
	    
	    System.out.println("PubSubService.publish(msg): discarding message because its payload is too long ("
			       + buffer.length 
			       + " > max="
			       + max_payload 
			       + ")");
	    return;
	}
	
	// If lifetime is specified, adjust deadline based on lifetime
	long lifetime = desc.getLifetime();
	if (lifetime != -1) {
	    Date deadline = desc.getDeadline();
	    if (deadline == null) {
		// If deadline attribute is already specified, we
		// assume it's correct. Otherwise we compute deadline
		// based on lifetime
		long now = System.currentTimeMillis();
		desc.setDeadline(new Date(now + lifetime));
	    }
	    // In any case we remove the lifetime attribute from the
	    // descriptor
	    desc.removeAttribute("_ltime");
	}
	
	casa.dodwan.util.Logger.log("pubsub", 
				    "> publish(id=" 
				    + desc.getDocumentId() + ")");
	
	try {
	    MessageSAR.publish(desc, buffer, 
			       DodwanEnvironment.getInstance().fragmentation_threshold,
			       cacheService_.cache);
 
	    eventDispatcher_.publish(desc);
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    
     // ------------------------------------------------------------
     /**
     * Publishes a message. Does nothing if the size of the payload of
     * the specified message is greater than the maximum payload size
     * (see {@link DodwanEnvironment}).
     * @param desc the descriptor of the message to be published.
     * @param file the file that contains the payload of the message to be published.
     * @throws Exception if the publication fails.
     */
    public void publish(Descriptor desc, File file) throws Exception {

	if (verbosity.isEnabled())
	    System.out.println("PubSubService.publish()");
	
	if (isSuspended()) {
	    if (verbosity.isEnabled())
		System.out.println("  This service is suspended. Aborting...");
	    return;
	}
	
	if (! file.exists()) {
	    System.out.println("PubSubService.publish(desc, file): file does not exist");
	    return;
	}

	int size = (int) file.length();

	int max_payload = DodwanEnvironment.getInstance().max_payload_size;
	if ((max_payload > 0) 
	    && (size > max_payload)) {
	    
	    System.out.println("PubSubService.publish(msg): discarding message because its payload is too long ("
			       + size 
			       + " > max="
			       + max_payload 
			       + ")");
	    return;
	}
	
	// If lifetime is specified, adjust deadline based on lifetime
	long lifetime = desc.getLifetime();
	if (lifetime != -1) {
	    Date deadline = desc.getDeadline();
	    if (deadline == null) {
		// If deadline attribute is already specified, we
		// assume it's correct. Otherwise we compute deadline
		// based on lifetime
		long now = System.currentTimeMillis();
		desc.setDeadline(new Date(now + lifetime));
	    }
	    desc.removeAttribute("_ltime");
	}
	
	casa.dodwan.util.Logger.log("pubsub", 
				    "> publish(id=" 
				    + desc.getDocumentId() + ")");
	
	// FG: on pourrait sans doute eviter de reassembler
	// localement les messages publies localement, en faisant en
	// sorte d'invoquer directement les Processors dont les
	// patterns matchent.

	try {
	    MessageSAR.publish(desc, file, 
			       DodwanEnvironment.getInstance().fragmentation_threshold,
			       cacheService_.cache);

	    eventDispatcher_.publish(desc);
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    
//      // ------------------------------------------------------------
//      /**
//       * @undocumented
//       *
//       * Cancels the propagation of a message
//       * @param descriptor
//       */
//     public void cancel(Descriptor descriptor) {
// 	 cacheService_.canceller.setPattern(descriptor);
//       }

      // ------------------------------------------------------------
      /**
       * @undocumented
       */
      public Console console() {
	  if (console_ == null)
	      console_ = new PubSubServiceConsole(this);
	return console_;
      }
    
    // ------------------------------------------------------------
     /**
     * Returns the keys of the registered subscriptions.
     * @return the set of the registered keys.
     */
    public Set<String> getKeys() {
	
	return map_.keySet();
    }
    
    // ------------------------------------------------------------
      /**
     * Returns the keys of the registered subscriptions that match a pattern.
     * @param pattern a descriptor that serves as a pattern for descriptor matching
     * @return the set of the keys that identify a registered subscription whose 
               descriptor matches the specified pattern.
     */
   public Set<String> getKeys(Descriptor pattern) {
	
	Set<String> result = new HashSet<String>();
	
	for (String key : map_.keySet()) {
	    
	    MessageReceiver recv = getReceiver(key);
	    if (recv.getPattern().matches(pattern))
		result.add(key);
	}
	
	return result;
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public MessageReceiver getReceiver(String key) {
	
	return map_.get(key);
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Collection<MessageReceiver> getReceivers() {
	
	return map_.values();
    }
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Collection<MessageReceiver> getReceivers(Descriptor pattern) {
	
	Set<MessageReceiver> result = new HashSet<MessageReceiver>();
	
	for (MessageReceiver recv : map_.values()) {
	    
	    if (recv.getPattern().matches(pattern))
		result.add(recv);
	}
	
	return result;
    }

    // ------------------------------------------------------------
    private void checkPayloadAvailability(String key) {

	Descriptor desc = cacheService_.cache.getDescriptor(key);
	if (desc == null) {
	    casa.dodwan.util.Logger.log("pubsub",
					"warning: unknown mid="
					+ key);
	    return;
	}

	int nbFrags = desc.getNbFragments();
	if (nbFrags <= 0)
	    casa.dodwan.util.Logger.log("pubsub",
					"warning: fragments unavailable for mid="
					+ key);
    }
    
    // ------------------------------------------------------------
    /**
     * Extract in memory the payload of a message that is present in the cache.
     * The message is identified by the key attribute stored in its descriptor 
     * (see method {@link casa.dodwan.docware.Descriptor#getKey()}).
     * @param key the key of the message
     * @return the byte array containing the payload of the message.
     * @throws Exception if the extraction fails.
     */
    public byte[] getAsBytes(String key)
    		throws Exception {

	checkPayloadAvailability(key);

	// casa.dodwan.util.Logger.log("pubsub", 
	// 			    "debug: getAsBytes(" + key + ")");
    	return MessageSAR.getAsBytes(key, cacheService_.cache);
    }

    // ------------------------------------------------------------
    /**
     * Extract in a file the payload of a message that is present in the cache.
     * The message is identified by the key attribute stored in its descriptor 
     * (see method {@link casa.dodwan.docware.Descriptor#getKey()}).
     * The file should be writeable.
     * @param key the key of the message.
     * @param file The file in which the payload of the message will be written.
     * @throws Exception if the extraction fails.
     */
    public void getAsFile(String key, File file)
    		throws Exception {

	checkPayloadAvailability(key);

	// casa.dodwan.util.Logger.log("pubsub", 
	// 			    "debug: getAsFile(" + key + "," + file + ")");
     	MessageSAR.getAsFile(key, file, cacheService_.cache);
    }

    // ------------------------------------------------------------
    public void cancelMessage(String key) {

	casa.dodwan.util.Logger.log("pubsub", "cancel(id=" + key + ")");
	try {
	    cacheService_.cache.cancel(key);
	} 
	catch (Exception e) {
	    casa.dodwan.util.Logger.log("pubsub", 
				    "cancel(" + key + ") failed");
	}
    }

    // ------------------------------------------------------------
    /**
      * @undocumented
      *
      * Resumes this PubSub service. 
      * Indicates that this service is restarted after a suspend.
      * Presently nothing is done but printing out logging information 
      * according to the verbosity level.
      */
   @Override
   public void resume() {
       
	if (verbosity.isEnabled())
	    System.out.println("PubSubService.resume()");
	
	if (isEnabled()) {
	    super.resume();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
     /**
      * Sets the verbosity level. 
      * @param level the level of verbosity. A value of 0 means that
      *        no information messages are printed out. A value
      *        different from 0 means that information messages are
      *        printed out on standard output.
      */
    public void setVerbosity(int level) {
	
	switch (level) {
	case 0:
	    verbosity.disable();
	    break;
	default:
	    verbosity.enable();
	    break;
	}
	
	System.out.println("PubSubService.setVerbosity(" + level + ")");
    }
    
    // ------------------------------------------------------------
     /**
      * @undocumented
      *
      * Suspends this PubSub service. 
      * Indicates that this service is temporarily suspended. It is typically  
      * called upon external events (host suspend to ram or suspend to disk).
      * Presently nothing is done but printing out logging information 
      * according to the verbosity level.
      */
    @Override
	public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("PubSubService.suspend()");
	
	if (isEnabled()) {
	    super.suspend();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
     /**
     * @undocumented
     */
    public String toString() {

	return "PubSubService"
	    + "\n nbProcessors: " + String.valueOf(map_.size())
	    + "\n isEnabled;" + String.valueOf(isEnabled())
	    + "\n isSuspended: " + String.valueOf(isSuspended());
    }

    // ============================================================
    // Implementation of the CacheListener interface
    // ============================================================

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void addDescriptor(Descriptor desc) {}

   // ------------------------------------------------------------
   /**
    * @undocumented
    */
    public void addPayload(String key, int fragId) {}

   // ------------------------------------------------------------
   /**
    * @undocumented
    */
    public void messageComplete(Descriptor desc) {
	
	String key = desc.getDocumentId();

	if (verbosity.isEnabled())
	    System.out.println("PubSubService.messageComplete('"
			       + key + "')");
	
	if (isSuspended()) {
	    if (verbosity.isEnabled())
		System.out.println("  This service is suspended. Aborting...");
	    return;
	}

	// FG: filtrage de doublons probablement inutile...
	if ((history_ != null)
	    && (history_.contains(key))) {
	    if (verbosity.isEnabled())
		System.out.println("PubSubService.messageComplete(): ignoring document " 
				   + key 
				   + " that has already been received in the past");
	    return;
	}

	casa.dodwan.util.Logger.log("pubsub", 
				    "< message(id=" + key + ")");

	history_.newMessageReceived(desc);
	// casa.dodwan.util.Logger.log("pubsub", 
	// 			    "debug: Calling receivers");
	
	receivers_.write(desc);
   }

   // ------------------------------------------------------------
   /**
    * @undocumented
    */
    public void cancel(String key) {}

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void remove(String key) {}

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void update(Descriptor desc) {}

}
