/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.run;

import java.net.ServerSocket;
import java.net.Socket;
import java.net.InetSocketAddress;
import java.io.File;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.lang.Runtime;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.net.Stream_IO;
import casa.dodwan.net.browseInterfaces;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.DataFormatting;

public class MessageReceiver extends Thread {

    private ServerSocket server_;
    private String key_;
    private File dir_;
    private String pattern_;
    private MessageBufferizer bufferizer_ = new MessageBufferizer();
    private InetSocketAddress localAddr_;
    private InetSocketAddress remoteAddr_ = null;
    private String fifo_ = null;
    
    // ==================================================
    private class ShutdownHook
	extends Thread {

	private MessageReceiver receiver_;
	
	public ShutdownHook(MessageReceiver receiver) {
	    receiver_ = receiver;
	}
	
	public void run() {

	    // System.out.println("Running ShutdownHook");
	    receiver_.removeSubscription();
	}
    }

    // ==================================================
    private class ClientHandler
	extends Thread {

	private Socket socket_;
	private File dir_;
	
	// --------------------------------------------------
	public ClientHandler(Socket socket, File dir) {
	    socket_ = socket;
	    dir_ = dir;
	}

	// --------------------------------------------------
	public void run() {

	    // Receiving message from input stream
	    byte buffer[] = null;
	    Message msg = null;
	    Descriptor desc = null;
	    byte[] payload = null;
	    try {
		// FG: il serait préférable de ne pas passer par un
		// buffer (potentiellement gigantesque), mais de
		// fonctionner plutôt en flux, au moins pour la partie
		// payload
		buffer = Stream_IO.receiveBundle(socket_.getInputStream());
		socket_.close();
		msg = bufferizer_.fromBuffer(buffer);
		desc = msg.getDescriptor();
		payload = msg.getPayload();
	    }
	    catch (Exception e) {
		System.out.println("Failed to receive message");
		try { socket_.close(); } catch (Exception f) {}
	    }
	    
	    // System.out.println("desc=" + desc);
	    // System.out.println("payload=<" + new String(payload) + ">");

	    // Extracting attributes from descriptor
	    String src = desc.getAttribute("src");
	    String key = desc.getDocumentId();
	    String fname = desc.getAttribute("file");

	    // Processing message
	    if (src == null) {
		System.out.println("Discard: unknown source");
		return;
	    }
	    System.out.println("Received message " + key + " from " + src);
	    String target = null;
	    if (fname == null) {
		target = src + "_" + key;
		System.out.println("This message does not contain a file. Saving in " + target);
	    }
	    else {
		target = src + "_" + fname;
		System.out.println("This message contains a file named " + fname
				   + ", saving in " + target);
	    }

	    // Saving message payload in local file
	    try {
		File file = new File(dir_, target);
		PrintStream out = new PrintStream(file);
		out.write(payload);
		out.flush();
		out.close();
		// System.out.println("Message payload saved in " + target);
	    }
	    catch (Exception e) {
		System.out.println("Failed to write " + target);
		try { socket_.close(); } catch (Exception f) {}
	    }
	}
    }

    // ==================================================
    public MessageReceiver(String[] args) throws Exception {

	if ((args.length == 0) || args[0].equals("-h")) {
	    help();
	}

	String path = null;
	try {
	    path = ArgumentParsing.searchString("-d", args);
	    pattern_ = ArgumentParsing.searchString("-p", args);
	} catch (ArgumentNotFoundException e) {
	    System.out.println(e);
	    System.exit(1);
	} catch (ArgumentParsingException e) {
	    System.out.println(e);
	    System.exit(1);
	}

	try {
	    try {
		String addr = ArgumentParsing.searchString("-a", args);
		remoteAddr_ = DataFormatting.toInetSocketAddress(addr);
	    } catch (ArgumentNotFoundException e) {}
	    try {
		fifo_ = ArgumentParsing.searchString("-f", args);
	    } catch (ArgumentNotFoundException e) {}
	} catch (ArgumentParsingException e) {
	    System.out.println(e);
	    System.exit(1);
	}

	if ((fifo_ == null) && (remoteAddr_ == null)) {
	    System.out.println("Missing argument: neead at least -a or -f.");
	    System.exit(1);
	}
	
	dir_ = new File(path);
	if (! dir_.isDirectory()) {
	    System.out.println("Invalid directory " + path);
	    System.exit(1);
	}
	
	int lport = 0;
	try {
	    lport = ArgumentParsing.searchInt("-lp", args);
	} catch (ArgumentNotFoundException e) {}
	catch (ArgumentParsingException e) {
	    System.out.println(e);
	    System.exit(1);
	}
	
	server_ = new ServerSocket(lport);
	if (lport == 0)
	    lport = server_.getLocalPort();
	localAddr_ = new InetSocketAddress(browseInterfaces.getInet4Address(), lport);
	
	System.out.println("Now listening to port #" + lport);
	System.out.println("Files will be saved in " + dir_);

	key_ = "mr_" + (int)(Math.random() * 10000);
    }

    // ------------------------------------------------------------
    protected void removeSubscription() {

	System.out.println("Removing subscription " + key_ + " for pattern: " + pattern_);
	sendCommand("do ps rem " + key_);
    }

    // ------------------------------------------------------------
    protected void setSubscription() {

	System.out.println("Setting subscription " + key_ + " for pattern: " + pattern_);
	sendCommand("do ps add -k " + key_
		    + " -desc " + pattern_
		    + " -a " + DataFormatting.toString(localAddr_));
    }

    // ------------------------------------------------------------
    private void sendCommand(String cmd) {

	// System.out.println("sendCommand(" + remoteAddr_ + "): " + cmd);
	
	if (remoteAddr_ != null) {
	    try {
		Socket socket = new Socket(remoteAddr_.getAddress(), remoteAddr_.getPort());
		PrintWriter out = new PrintWriter(socket.getOutputStream());
		out.println(cmd);
		out.flush();
		socket.close();
	    }
	    catch (Exception e) {
		System.out.println("Failed to send command to " + remoteAddr_);
	    }
	    return;
	}

	if (fifo_ != null) {
	    try {
		PrintWriter out = new PrintWriter(fifo_);
		out.println(cmd);
		out.flush();
		out.close();
	    }
	    catch (Exception e) {
		System.out.println("Failed to write command in " + fifo_);
	    }
	    return;
	}
	
    }

    // ------------------------------------------------------------
    public void run() {

	Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));
	
	try {
	    setSubscription();
	    System.out.println();
	    
	    while (true) {
		Socket socket = server_.accept();
		ClientHandler receiver = new ClientHandler(socket, dir_);
		receiver.start();
	    }
	}
	catch (Exception e) {
		System.out.println("Closing socket " + server_);
		try { server_.close(); } catch (Exception f) {};
		try { removeSubscription(); } catch (Exception f) {};
		e.printStackTrace();
	    }
    }
	
    // ------------------------------------------------------------
    public void help() {

	System.out.println("Syntax: MessageReceiver [-d dir] [-p pattern] [-a  addr|-f fifo] {-lp local port}");
	System.exit(1);
    }

    // ------------------------------------------------------------
    public static void main(String[] args) throws Exception {

	MessageReceiver receiver = new MessageReceiver(args);
	receiver.run();
    }
	
}

