/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.run;

import java.util.Random;
import java.util.Date;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileReader;
import java.io.PrintStream;
import java.io.File;
import java.lang.Runtime;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.Command;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Console;
import casa.dodwan.util.TCPConsole;
import casa.dodwan.util.RandomSeed;
import casa.dodwan.util.DataFormatting;

public class dodwand extends Thread {

    // This class defines a wrapper for an instance of class Dodwan.

    
    // ==================================================
    private class TerminationThread
	extends Thread {

	private long wait_;
	private dodwand wrapper_;
	
	public TerminationThread(long wait, dodwand wrapper) {
	    wait_ = wait;
	    wrapper_ = wrapper;
	}
	
	public void run() {
	    try {
		if (wait_ <= 0)
		    return;

		sleep(wait_);
		System.out.println("Terminating as planned");
		wrapper_.dodwan.offline();
		wrapper_.clear();
		wrapper_.dodwan.doNotify();
	    } catch (Exception e) {}
	}
    }
    // ==================================================
    private class ShutdownHook
	extends Thread {

	private dodwand wrapper_;
	
	public ShutdownHook(dodwand wrapper) {
	    wrapper_ = wrapper;
	}
	
	public void run() {

	    try {
		System.out.println("Got a termination signal");
		wrapper_.dodwan.offline();
		wrapper_.clear();
		wrapper_.dodwan.doNotify(); 
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    // ==================================================

    public Dodwan dodwan   = null;
    private String ifile    = null;
    private String cfile    = null;
    private long begin      = -1;
    private long end        = -1;
    private long seed       = -1;
    private Console console = null;
    private File pid_file   = null;
    private File cport_file = null;
    private boolean cleared = false;
    
    // ------------------------------------------------------------
    public dodwand(String args[]) throws Exception {

	// casa.dodwan.util.Catalog.main(args);
	// System.exit(1);
	
	if (ArgumentParsing.searchIndex("-h", args) != -1) {
	    exit(0);
	}
	
	begin = -1;
	try {
	    begin = ArgumentParsing.searchLong("-begin", args);
	} catch (Exception e) {};
	    
	end = -1;
	try {
	    end = ArgumentParsing.searchLong("-end", args);
	} catch (Exception e) {};
	
	seed = System.currentTimeMillis();
	try {
		seed = ArgumentParsing.searchLong("-seed", args);
	} catch (Exception e) {};
	RandomSeed.setSeed(seed);
	
	cfile = null;
	try {
	    cfile = ArgumentParsing.searchString("-c", args);
	} catch (Exception e) {};

	ifile = null;
	try {
	    ifile = ArgumentParsing.searchString("-i", args);
	} catch (Exception e) {};

	
	DodwanEnvironment env = DodwanEnvironment.getInstance();
	env.displayProperties();
	pid_file = new File(env.directory, "pid");
	
	System.out.println("Instantiation of Dodwan on " + SystemEnvironment.host);
	
	dodwan = new Dodwan();
	
	console = new Console();
	console.addSubConsole("d", "[d]odwan", dodwan.console());
	
	int console_port = env.console_port;
	if (console_port != -1) {
	    TCPConsole mainConsole =
		new TCPConsole(console, console_port);
	    if (console_port == 0)
		console_port = mainConsole.getLocalPort();
	    System.out.println("A console is now listening to port TCP#"
			       + console_port);
	    cport_file = new File(env.directory, "cport");
	    PrintStream ps = new PrintStream(cport_file);
	    ps.println(console_port);
	    ps.close();
	}

	if (cfile != null) {
	    System.out.println("Parsing command file " + cfile);
	    console.parseCommandFile(cfile, System.out);
	}
    }

    // ------------------------------------------------------------
    public void run() {
	
	Runtime.getRuntime().addShutdownHook(new ShutdownHook(this));

	if (begin != -1) {
	    System.out.println("Deferring start until " + new Date(begin));
	    long now = System.currentTimeMillis();
	    long wait = begin - now;
	    if (wait > 0)
		try {
		    Thread.sleep(wait);
		}
		catch (Exception e) {}
	}
	
	long wait = -1;
	if (end != -1) {
	    long now = System.currentTimeMillis();
	    wait = end - now;
	    System.out.println("Will stop in " +
			       DataFormatting.to_hh_mm_ss(wait));
	    TerminationThread th = new TerminationThread(wait, this);
	    th.start();
	}

	// Switch DoDWAN online
	dodwan.online();

	// If option '-i' is specified, start a console, taking as
	// input either stdin or a FIFO file
	boolean interactive = (ifile != null);
	if (interactive) {
	    System.out.println("Input from " + ifile);
	    PrintStream out = System.out;
	    boolean fifo = (! ifile.equals("-"));	
	    try {
		do {
		    BufferedReader in = (fifo?
					 new BufferedReader(new FileReader(ifile)):
					 new BufferedReader(new InputStreamReader(System.in)));
		    console.parseCommandStream(in, out, true);
		} while (fifo);
	    }
	    catch (Exception e) {
		System.out.println("Failed to open " + ifile);
	    }
	}
    }

    // ------------------------------------------------------------
    public void clear() {

	if (cleared)
	    return;
	
	// Trying to remove pid file (assuming there is one)
	if (pid_file.exists()) {
	    try {
		pid_file.delete();
		System.out.println("File " + pid_file + " deleted");
		pid_file = null;
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}
	
	// Trying to remove cport file (assuming there is one)
	if (cport_file.exists()) {
	    try {
		cport_file.delete();
		System.out.println("File " + cport_file + " deleted");
		cport_file = null;
	    }
	    catch (Exception e) {
		e.printStackTrace();
	    }
	}

	cleared = true;
    }
    
    // ------------------------------------------------------------
    public static void exit(int error) {
	help();
	Command.exit(error);
    }

    // ------------------------------------------------------------
    public static void help() {
	System.out.println("\n"
			   + "Syntax: dodwand [-h] [-c file] [-i (-|fifo file)] [-seed value] [-begin time (ms)] [-end time (ms)] -Dkey=value ...");
	System.out.println("  -h: help");
	System.out.println("  -c: command file (read once at startup)");
	System.out.println("  -i: interactive mode, with a console taking either stdin or a fifo file as input (default: no console)");
	System.out.println("  -seed:  seed for the random generator");
	System.out.println("  -begin: deferred start time (EPOCH in ms)");
	System.out.println("  -end:   termination time (EPOCH in ms)");
    }

    // ------------------------------------------------------------
    public static void main(String[] args) throws Exception {

	dodwand wrapper = new dodwand(args);
	wrapper.start();

	// Wait until Dodwan is shut down (either via a 'do shutdown'
	// issued in a console, or via the builtin TerminationThread
	// if end time is specified)
	wrapper.dodwan.doWait();
	
	// Try to terminate DoDWAN gracefully
	wrapper.clear();
	System.out.println("DoDWAN terminated");
	System.out.flush();
	
	System.exit(0);
    }

}

