/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class Console {

    private class SubConsole {

	protected String prefix_;
	protected Console console_;
    }

    protected String prompt_ = "% ";

    private boolean quitAllowed_ = false;

    protected Map<String, SubConsole> subconsoles_ 
	= new HashMap<String, SubConsole>();

    // ------------------------------------------------------------
    public void help(String offset, PrintStream out) {
	
	out.println(offset   + "[h]elp\n"
		    + offset + "[source] <cmdFile>\n"
		    + offset + "[cron] <-in hh:mm:ss> <-every hh:mm:ss> <-cmd command>\n"
		    + (quitAllowed_?offset + "[quit]\n":""));

	subHelp(offset, out);
    }


    // ------------------------------------------------------------
    public void subHelp(String offset, PrintStream out) {
	for (SubConsole sc : subconsoles_.values()) {
//  	    out.println("sc= " + sc
//  			+ ", console=" + sc.console_
//  			+ ", offset=" + offset
//  			+ ", prefix=" + sc.prefix_
//  			+ ", out=" + out);

	    sc.console_.help(offset + sc.prefix_ + " ", out);
	    out.println();
	}
    }

    // ------------------------------------------------------------
    public void quitAllowed(boolean allowed) {
	quitAllowed_ = allowed;
    }

    // ------------------------------------------------------------
    public void source(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("", out);
	    return;
	}
	
	for (int i=1; i<arguments.length; i++)
	    parseCommandFile(arguments[i], out);
    }
    
    // ------------------------------------------------------------
    public void addSubConsole(String key, String prefix, 
			      Console console) {

//  	System.out.println("Adding subconsole with key=" + key
//  			   + " in " + this);

	SubConsole sc = new SubConsole();
	sc.prefix_ = prefix;
	sc.console_ = console;

	subconsoles_.put(key, sc);
    }

    // ------------------------------------------------------------
    public void removeSubConsole(String key) {

	subconsoles_.remove(key);
    }

    // ------------------------------------------------------------
    public void parseCommandFile(String fname, PrintStream out) {
	
	BufferedReader fin;
	try {
	    fin = new BufferedReader(new FileReader(fname));
	} catch (FileNotFoundException e) {
	    out.println("Console.parseCommandFile(" + fname
			+ ") failed: file not found.");
	    return;
	}
	
	// out.println("--- Parsing command file '" + fname + "' ---");
	
	try {
	    String line = fin.readLine();
	    
	    while (line != null) {
		parseCommandLine(line, fin, out, true);
		line = fin.readLine();
	    }
	} catch (Exception e) {
	    out.println("Console.parseCommandFile(): " + e);
	}
	
	// out.println("--- End of command file ---");
    }

    // ------------------------------------------------------------
    private final String commandRegex = "\\s*(" +
	"(([^\"\\[\\]\\s]+))" +
	"|" +
	"(\\[([^\\]]+)\\])" +
	"|" +
	"(\\\"([^\"]+)\\\")" +
	")\\s*";
    
    // ------------------------------------------------------------
    public void parseCommandLine(String line, BufferedReader in,
				 PrintStream out, boolean echo) 
	throws Exception {
	
	if (line == null)
	    return;

	String[] cmds = line.split(";");

	for (String cmd: cmds) {
	    cmd = cmd.trim();
	    if (echo) out.println(prompt_ + cmd);
	    parseCommandString(cmd, in, out);
	}
    }

    // ------------------------------------------------------------
    public void parseCommandString(String command, BufferedReader in,
				   PrintStream out) 
	throws Exception {
	
	if (command == null)
	    return;

	Pattern p = Pattern.compile(commandRegex);
	Matcher m = p.matcher(command);
	Vector<String> args = new Vector<String>(); 
	
	while (m.find())
	    args.add(m.group(1).replaceAll("\"", ""));
	
	String[] arguments = new String[args.size()];
	Iterator<String> it = args.iterator();
	int i = 0;
	while (it.hasNext())
	    arguments[i++] = it.next();
	
	processCommand(arguments, in, out);
    }
    
    // ------------------------------------------------------------
    public void parseCommandStream(BufferedReader in,
				   PrintStream out,
				   boolean echo) {
	
	try {
	    if (! echo) out.print(prompt_);
	    String line = in.readLine();
	    while ((line != null) && 
		   (! ( quitAllowed_ && line.startsWith("quit")))) {
		parseCommandLine(line, in, out, echo);
		
		if (! echo) out.print(prompt_);
		line = in.readLine();
	    }
	} catch (Exception e) {
	    out.println("Console.parseCommandStream()");
	    e.printStackTrace();
	}
    }
    
    // ------------------------------------------------------------
    public void processCommand(String[] arguments, BufferedReader in,
			       PrintStream out) {

	if (arguments.length == 0)
	    return;

	String prefix = arguments[0];
	
	// Discard empty lines and comment lines
	if ((arguments.length == 0) || (prefix.startsWith("#")))
	    return;

	// Process help command
	if (arguments[0].startsWith("h")) {
	    help("", out);
	    return;
	}
	
	// Process command file
	if (arguments[0].startsWith("source")) {
	    source(arguments, out);
	    return;
	}
	
	// Process deferred (cron) commands
	if (arguments[0].startsWith("cron")) {
	    processCronCommand(arguments, in, out);
	    return;
	}

	// Check for commands issued to subconsoles
	for (String key : subconsoles_.keySet()) {
	    
//  	    System.out.println("Checking if prefix '"
//  			       + prefix + "' matches key '"
//  			       + key + "'");

	    // If a subconsole exists whose key matches the prefix of the
	    // command, then pass all arguments of this command to this
	    // subconsole and returns
	    if (prefix.startsWith(key)) {

		String[] subargs = new String[arguments.length - 1];
		for (int j = 1; j < arguments.length; j++)
		    subargs[j-1] = arguments[j];

		SubConsole sc = subconsoles_.get(key);
		sc.console_.processCommand(subargs, in, out);
		return;
	    }
	}

	out.println("Unknown command. Try 'help'");
    }

    // ------------------------------------------------------------
    public void processCronCommand(String[] arguments, BufferedReader in,
				   PrintStream out) {
	
	if (arguments.length < 3) {
	    help("", out);
	    return;
	}
	
	long delay = -1;
	long period = -1;
	String[] taskArguments;
	try {
	    try {
		delay = ArgumentParsing.search_hh_mm_ss("-in", arguments);
	    }
	    catch (ArgumentNotFoundException e) {
	    }
	    
	    try {
		period = ArgumentParsing.search_hh_mm_ss("-every", arguments);
	    }
	    catch (ArgumentNotFoundException e) {
	    }
	    
	    taskArguments = ArgumentParsing
		.searchStringArray("-cmd", arguments);
	}
	catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	}
	catch (Exception e) {
	    out.println(e);
	    return;
	}
	
	if (delay + period == -2) {
	    out.println("Arguments missing in cron command");
	    help("", out);
	    return;
	}
	
	ConsoleTask task = new ConsoleTask(this, taskArguments, in, out);
	
	if (delay == -1)
	    delay = (long) (new Random().nextDouble() * period);
	
	out.print("The command: ' ");
	for (int i = 0; i < taskArguments.length; i++)
	    out.print(taskArguments[i] + " ");
	out
	    .println("' will be executed in "
		     + (delay / 1000)
		     + " seconds"
		     + ((period != -1) ? (" and repeated every "
					  + (period / 1000) + " seconds") : ""));
	
	if (period == -1)
	    Timer.getInstance().schedule(task, delay);
	else
	    Timer.getInstance().schedule(task, delay, period);
    }
    

}
